@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2012 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -43,23 +43,13 @@ int main(int argc, char *argv[]) {
 
     if(WSAStartup(MAKEWORD(1, 1), &wsa_state))
         return 1;
-    main_initialize(argc>1 ? argv[1] : NULL, argc>2 ? argv[2] : NULL);
+    main_initialize();
+    if(main_configure(argc>1 ? argv[1] : NULL, argc>2 ? argv[2] : NULL))
+        return 1;
     main_execute();
     return 0;
 }
 
-void win_log(LPSTR line) { /* also used in log.c */
-    LPTSTR tstr;
-
-    tstr=str2tstr(line);
-    RETAILMSG(TRUE, (TEXT("%s\r\n"), tstr));
-    free(tstr);
-}
-
-void exit_win32(int exit_code) {
-    exit(exit_code);
-}
-
 int passwd_cb(char *buf, int size, int rwflag, void *userdata) {
     return 0; /* not implemented */
 }
