@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2012 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -47,20 +47,28 @@
 #endif /* va_copy */
 
 static u8 canary[10]; /* 80-bit canary value */
+static volatile int canary_initialized=0;
 
-typedef struct alloc_list ALLOC_LIST;
+typedef struct alloc_list_struct ALLOC_LIST;
 
-typedef struct alloc_tls {
+typedef struct {
     ALLOC_LIST *head;
     size_t bytes, blocks;
 } ALLOC_TLS;
 
-struct alloc_list {
+struct alloc_list_struct {
     ALLOC_LIST *prev, *next;
     ALLOC_TLS *tls;
     size_t size;
+    int valid_canary;
     unsigned int magic;
+        /* at least on IA64 allocations need to be aligned */
+#ifdef __GNUC__
+} __attribute__((aligned(16)));
+#else
+    int padding[2]; /* the number of integers is architecture-specific */
 };
+#endif
 
 static void set_alloc_tls(ALLOC_TLS *);
 static ALLOC_TLS *get_alloc_tls();
@@ -70,8 +78,7 @@ char *str_dup(const char *str) {
     char *retval;
 
     retval=str_alloc(strlen(str)+1);
-    if(retval)
-        strcpy(retval, str);
+    strcpy(retval, str);
     return retval;
 }
 
@@ -91,8 +98,6 @@ char *str_vprintf(const char *format, va_list start_ap) {
     va_list ap;
 
     p=str_alloc(size);
-    if(!p)
-        return NULL;
     for(;;) {
         va_copy(ap, start_ap);
         n=vsnprintf(p, size, format, ap);
@@ -103,10 +108,6 @@ char *str_vprintf(const char *format, va_list start_ap) {
         else          /* glibc 2.0, WIN32, etc. */
             size*=2;  /* twice the old size */
         np=str_realloc(p, size);
-        if(!np) {
-            str_free(p);
-            return NULL;
-        }
         p=np; /* LOL */
     }
 }
@@ -136,19 +137,17 @@ static ALLOC_TLS *get_alloc_tls() {
 
 #ifdef USE_FORK
 
-static ALLOC_TLS *global_tls;
-static pid_t previous_pid=0;
+static ALLOC_TLS *global_tls=NULL;
 
 void str_init() {
 }
 
 static void set_alloc_tls(ALLOC_TLS *tls) {
     global_tls=tls;
-    previous_pid=getpid();
 }
 
 static ALLOC_TLS *get_alloc_tls() {
-    return previous_pid==getpid() ? global_tls : NULL;
+    return global_tls;
 }
 
 #endif /* USE_FORK */
@@ -177,43 +176,23 @@ static DWORD tls_index;
 
 void str_init() {
     tls_index=TlsAlloc();
-    if(tls_index==TLS_OUT_OF_INDEXES) {
-        s_log(LOG_ERR, "TlsAlloc failed");
-        die(1);
-    }
 }
 
 static void set_alloc_tls(ALLOC_TLS *alloc_tls) {
-    if(!TlsSetValue(tls_index, alloc_tls)) {
-        s_log(LOG_ERR, "TlsSetValue failed");
-        die(1);
-    }
+    TlsSetValue(tls_index, alloc_tls);
 }
 
 static ALLOC_TLS *get_alloc_tls() {
-    ALLOC_TLS *alloc_tls;
-
-    alloc_tls=TlsGetValue(tls_index);
-    if(!alloc_tls && GetLastError()!=ERROR_SUCCESS) {
-        s_log(LOG_ERR, "TlsGetValue failed");
-        die(1);
-    }
-    return alloc_tls;
+    return TlsGetValue(tls_index);
 }
 
 #endif /* USE_WIN32 */
 
-void str_canary() {
-    ALLOC_TLS *alloc_tls;
-    ALLOC_LIST *tmp;
-
+void str_canary_init() {
+    if(canary_initialized) /* prevent double initialization on config reload */
+        return;
     RAND_bytes(canary, sizeof canary);
-    /* str_canary() is executed with only one thread,
-       so it is enough to process only one list */
-    alloc_tls=get_alloc_tls();
-    if(alloc_tls) /* need to rewrite existing canaries */
-        for(tmp=alloc_tls->head; tmp; tmp=tmp->next)
-            memcpy((u8 *)(tmp+1)+tmp->size, canary, sizeof canary);
+    canary_initialized=1; /* after RAND_bytes */
 }
 
 void str_cleanup() {
@@ -221,8 +200,8 @@ void str_cleanup() {
 
     alloc_tls=get_alloc_tls();
     if(alloc_tls) {
-        while(alloc_tls->head)
-            str_free(alloc_tls->head+1);
+        while(alloc_tls->head) /* str_free macro requires lvalue parameter */
+            str_free_debug(alloc_tls->head+1, __FILE__, __LINE__);
         set_alloc_tls(NULL);
         free(alloc_tls);
     }
@@ -232,15 +211,18 @@ void str_stats() {
     ALLOC_TLS *alloc_tls;
 
     alloc_tls=get_alloc_tls();
-    if(alloc_tls)
-        s_log(LOG_DEBUG,
-            "str_stats: %lu block(s), %lu data byte(s), %lu control byte(s)",
-            (unsigned long int)alloc_tls->blocks,
-            (unsigned long int)alloc_tls->bytes,
-            (unsigned long int)(alloc_tls->blocks*
-                (sizeof(ALLOC_LIST)+sizeof canary)));
-    else
+    if(!alloc_tls) {
         s_log(LOG_DEBUG, "str_stats: alloc_tls not initialized");
+        return;
+    }
+    if(!alloc_tls->blocks && !alloc_tls->bytes)
+        return; /* skip if no data is allocated */
+    s_log(LOG_DEBUG, "str_stats: %lu block(s), "
+            "%lu data byte(s), %lu control byte(s)",
+        (unsigned long int)alloc_tls->blocks,
+        (unsigned long int)alloc_tls->bytes,
+        (unsigned long int)(alloc_tls->blocks*
+            (sizeof(ALLOC_LIST)+sizeof canary)));
 }
 
 void *str_alloc_debug(size_t size, char *file, int line) {
@@ -251,25 +233,29 @@ void *str_alloc_debug(size_t size, char *file, int line) {
     if(!alloc_tls) { /* first allocation in this thread */
         alloc_tls=calloc(1, sizeof(ALLOC_TLS));
         if(!alloc_tls)
-            fatal("Out of memory", file, line);
+            fatal_debug("Out of memory", file, line);
         alloc_tls->head=NULL;
         alloc_tls->bytes=alloc_tls->blocks=0;
         set_alloc_tls(alloc_tls);
     }
     alloc_list=calloc(1, sizeof(ALLOC_LIST)+size+sizeof canary);
     if(!alloc_list)
-        fatal("Out of memory", file, line);
-    memcpy((u8 *)(alloc_list+1)+size, canary, sizeof canary);
+        fatal_debug("Out of memory", file, line);
+
     alloc_list->prev=NULL;
     alloc_list->next=alloc_tls->head;
     alloc_list->tls=alloc_tls;
     alloc_list->size=size;
+    alloc_list->valid_canary=canary_initialized; /* before memcpy */
+    memcpy((u8 *)(alloc_list+1)+size, canary, sizeof canary);
     alloc_list->magic=0xdeadbeef;
+
     if(alloc_tls->head)
         alloc_tls->head->prev=alloc_list;
     alloc_tls->head=alloc_list;
     alloc_tls->bytes+=size;
     alloc_tls->blocks++;
+
     return alloc_list+1;
 }
 
@@ -282,8 +268,7 @@ void *str_realloc_debug(void *ptr, size_t size, char *file, int line) {
     alloc_list=realloc(previous_alloc_list,
         sizeof(ALLOC_LIST)+size+sizeof canary);
     if(!alloc_list)
-        fatal("Out of memory", file, line);
-    memcpy((u8 *)(alloc_list+1)+size, canary, sizeof canary);
+        fatal_debug("Out of memory", file, line);
     if(alloc_list->tls) { /* not detached */
         /* refresh possibly invalidated linked list pointers */
         if(alloc_list->tls->head==previous_alloc_list)
@@ -296,6 +281,8 @@ void *str_realloc_debug(void *ptr, size_t size, char *file, int line) {
         alloc_list->tls->bytes+=size-alloc_list->size;
     }
     alloc_list->size=size;
+    alloc_list->valid_canary=canary_initialized; /* before memcpy */
+    memcpy((u8 *)(alloc_list+1)+size, canary, sizeof canary);
     return alloc_list+1;
 }
 
@@ -332,7 +319,7 @@ void str_free_debug(void *ptr, char *file, int line) {
         return;
     str_detach_debug(ptr, file, line);
     alloc_list=(ALLOC_LIST *)ptr-1;
-    alloc_list->magic=0xdefec8ed; /* to detect double free */
+    alloc_list->magic=0xdefec8ed; /* to detect double free attempts */
     free(alloc_list);
 }
 
@@ -340,12 +327,17 @@ static ALLOC_LIST *get_alloc_list_ptr(void *ptr, char *file, int line) {
     ALLOC_LIST *alloc_list;
 
     alloc_list=(ALLOC_LIST *)ptr-1;
-    if(alloc_list->magic!=0xdeadbeef) /* not allocated by str_alloc() */
-        fatal("Bad magic", file, line);
+    if(alloc_list->magic!=0xdeadbeef) { /* not allocated by str_alloc() */
+        if(alloc_list->magic==0xdefec8ed)
+            fatal_debug("Double free attempt", file, line);
+        else
+            fatal_debug("Bad magic", file, line); /* LOL */
+    }
     if(alloc_list->tls /* not detached */ && alloc_list->tls!=get_alloc_tls())
-        fatal("Wrong thread", file, line);
-    if(memcmp((u8 *)ptr+alloc_list->size, canary, sizeof canary))
-        fatal("Dead canary", file, line);
+        fatal_debug("Memory allocated in a different thread", file, line);
+    if(alloc_list->valid_canary &&
+            memcmp((u8 *)ptr+alloc_list->size, canary, sizeof canary))
+        fatal_debug("Dead canary", file, line); /* LOL */
     return alloc_list;
 }
 
